@@ -1,14 +1,13 @@
 import { of } from 'rxjs';
 
-import { DataQueryRequest, dateTime, LoadingState } from '@grafana/data';
+import { DataQueryError, DataQueryRequest, DataQueryResponse, dateTime, LoadingState } from '@grafana/data';
 import { config } from '@grafana/runtime';
 
-import { createLokiDatasource } from './__mocks__/datasource';
-import { getMockFrames } from './__mocks__/frames';
 import { LokiDatasource } from './datasource';
-import * as logsTimeSplit from './logsTimeSplitting';
-import * as metricTimeSplit from './metricTimeSplitting';
+import { createLokiDatasource } from './mocks/datasource';
+import { getMockFrames } from './mocks/frames';
 import { runSplitQuery } from './querySplitting';
+import { LOKI_MAX_QUERY_BYTES_READ_ERROR_MSG_PREFIX, LOKI_TIMEOUT_ERROR_MSG } from './responseUtils';
 import { trackGroupedQueries } from './tracking';
 import { LokiQuery, LokiQueryDirection, LokiQueryType } from './types';
 
@@ -18,6 +17,7 @@ jest.mock('uuid', () => ({
 }));
 
 const originalShardingFlagState = config.featureToggles.lokiShardSplitting;
+const originalLokiQueryLimitsContextState = config.featureToggles.lokiQueryLimitsContext;
 const originalErr = console.error;
 beforeEach(() => {
   jest.spyOn(console, 'error').mockImplementation(() => {});
@@ -28,22 +28,23 @@ beforeAll(() => {
     callback();
   });
   config.featureToggles.lokiShardSplitting = false;
+  config.featureToggles.lokiQueryLimitsContext = true;
 });
 afterAll(() => {
   jest.mocked(global.setTimeout).mockReset();
   config.featureToggles.lokiShardSplitting = originalShardingFlagState;
+  config.featureToggles.lokiQueryLimitsContext = originalLokiQueryLimitsContextState;
   console.error = originalErr;
 });
 
 describe('runSplitQuery()', () => {
   let datasource: LokiDatasource;
+  const from = dateTime('2023-02-08T05:00:00.000Z');
+  const to = dateTime('2023-02-10T06:00:00.000Z');
   const range = {
-    from: dateTime('2023-02-08T05:00:00.000Z'),
-    to: dateTime('2023-02-10T06:00:00.000Z'),
-    raw: {
-      from: dateTime('2023-02-08T05:00:00.000Z'),
-      to: dateTime('2023-02-10T06:00:00.000Z'),
-    },
+    from,
+    to,
+    raw: { from, to },
   };
 
   const createRequest = (targets: LokiQuery[], overrides?: Partial<DataQueryRequest<LokiQuery>>) => {
@@ -72,6 +73,21 @@ describe('runSplitQuery()', () => {
     });
   });
 
+  test('Interpolates queries before execution', async () => {
+    const request = createRequest([{ expr: 'count_over_time({a="b"}[$__auto])', refId: 'A', step: '$step' }]);
+    datasource = createLokiDatasource({
+      replace: (input = '') => {
+        return input.replace('$__auto', '5m').replace('$step', '5m');
+      },
+      getVariables: () => [],
+    });
+    jest.spyOn(datasource, 'runQuery').mockReturnValue(of({ data: [] }));
+    await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
+      expect(jest.mocked(datasource.runQuery).mock.calls[0][0].targets[0].expr).toBe('count_over_time({a="b"}[5m])');
+      expect(jest.mocked(datasource.runQuery).mock.calls[0][0].targets[0].step).toBe('5m');
+    });
+  });
+
   test('Skips partial updates as an option', async () => {
     await expect(runSplitQuery(datasource, request, { skipPartialUpdates: true })).toEmitValuesWith((emitted) => {
       // 3 days, 3 chunks, 3 requests.
@@ -81,20 +97,59 @@ describe('runSplitQuery()', () => {
     });
   });
 
-  test('Retries retriable failed requests', async () => {
+  test('Retries retriable failed requests (timeout)', async () => {
     jest
       .mocked(datasource.runQuery)
-      .mockReturnValueOnce(of({ state: LoadingState.Error, errors: [{ refId: 'A', message: 'timeout' }], data: [] }));
-    await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-      // 3 days, 3 chunks, 1 retry, 4 requests.
+      .mockReturnValueOnce(
+        of({ state: LoadingState.Error, errors: [{ refId: 'A', message: LOKI_TIMEOUT_ERROR_MSG }], data: [] })
+      );
+    await expect(runSplitQuery(datasource, request)).toEmitValuesWith((response) => {
+      expect(response[0].state).toBe(LoadingState.Done);
+      // 3 days, 3 chunks, 1 retry = 4 requests.
       expect(datasource.runQuery).toHaveBeenCalledTimes(4);
     });
   });
 
+  test('Does not retry max bytes errors (error on initial)', async () => {
+    jest.mocked(datasource.runQuery).mockReturnValueOnce(
+      of({
+        state: LoadingState.Error,
+        errors: [{ refId: 'A', message: LOKI_MAX_QUERY_BYTES_READ_ERROR_MSG_PREFIX }],
+        data: [],
+      })
+    );
+    await expect(runSplitQuery(datasource, request)).toEmitValuesWith((response) => {
+      expect(response[0].state).toBe(LoadingState.Error);
+      // Stop immediately
+      expect(datasource.runQuery).toHaveBeenCalledTimes(1);
+    });
+  });
+
+  test('Does not retry max bytes errors (error on subsequent)', async () => {
+    const { logFrameA } = getMockFrames();
+    jest
+      .mocked(datasource.runQuery)
+      .mockReturnValueOnce(of({ data: [logFrameA], refId: 'A' }))
+      .mockReturnValueOnce(
+        of({
+          state: LoadingState.Error,
+          errors: [{ refId: 'A', message: LOKI_MAX_QUERY_BYTES_READ_ERROR_MSG_PREFIX }],
+          data: [],
+        })
+      );
+    await expect(runSplitQuery(datasource, request)).toEmitValuesWith((response) => {
+      expect(response[0].state).toBe(LoadingState.Error);
+      // Stop immediately after error
+      expect(datasource.runQuery).toHaveBeenCalledTimes(2);
+    });
+  });
+
   test('Does not retry failed queries as an option', async () => {
     jest
       .mocked(datasource.runQuery)
-      .mockReturnValueOnce(of({ state: LoadingState.Error, errors: [{ refId: 'A', message: 'timeout' }], data: [] }));
+      .mockReturnValueOnce(
+        of({ state: LoadingState.Error, errors: [{ refId: 'A', message: LOKI_TIMEOUT_ERROR_MSG }], data: [] })
+      );
     await expect(runSplitQuery(datasource, request, { disableRetry: true })).toEmitValuesWith(() => {
       // No retries
       expect(datasource.runQuery).toHaveBeenCalledTimes(1);
@@ -152,6 +207,19 @@ describe('runSplitQuery()', () => {
               _i: 1676008800000,
             }),
           }),
+          targets: [
+            {
+              expr: 'count_over_time({a="b"}[1m])',
+              legendFormat: undefined,
+              refId: 'A',
+              step: undefined,
+              limitsContext: {
+                expr: 'count_over_time({a="b"}[1m])',
+                from: from.valueOf(),
+                to: to.valueOf(),
+              },
+            },
+          ],
         })
       );
 
@@ -170,6 +238,15 @@ describe('runSplitQuery()', () => {
               _i: 1676005140000,
             }),
           }),
+          targets: [
+            {
+              expr: 'count_over_time({a="b"}[1m])',
+              legendFormat: undefined,
+              refId: 'A',
+              step: undefined,
+              limitsContext: undefined,
+            },
+          ],
         })
       );
 
@@ -188,6 +265,15 @@ describe('runSplitQuery()', () => {
               _i: 1675918740000,
             }),
           }),
+          targets: [
+            {
+              expr: 'count_over_time({a="b"}[1m])',
+              legendFormat: undefined,
+              refId: 'A',
+              step: undefined,
+              limitsContext: undefined,
+            },
+          ],
         })
       );
     });
@@ -212,6 +298,19 @@ describe('runSplitQuery()', () => {
               _i: 1676008800000,
             }),
           }),
+          targets: [
+            {
+              expr: 'count_over_time({a="b"}[1m])',
+              legendFormat: undefined,
+              refId: 'A',
+              step: '10s',
+              limitsContext: {
+                expr: 'count_over_time({a="b"}[1m])',
+                from: from.valueOf(),
+                to: to.valueOf(),
+              },
+            },
+          ],
         })
       );
 
@@ -230,6 +329,15 @@ describe('runSplitQuery()', () => {
               _i: 1676005190000,
             }),
           }),
+          targets: [
+            {
+              expr: 'count_over_time({a="b"}[1m])',
+              legendFormat: undefined,
+              refId: 'A',
+              step: '10s',
+              limitsContext: undefined,
+            },
+          ],
         })
       );
 
@@ -248,37 +356,97 @@ describe('runSplitQuery()', () => {
               _i: 1675918790000,
             }),
           }),
+          targets: [
+            {
+              expr: 'count_over_time({a="b"}[1m])',
+              legendFormat: undefined,
+              refId: 'A',
+              step: '10s',
+              limitsContext: undefined,
+            },
+          ],
         })
       );
     });
   });
 
-  test('Handles and reports errors', async () => {
-    jest
-      .spyOn(datasource, 'runQuery')
-      .mockReturnValue(of({ state: LoadingState.Error, error: { refId: 'A', message: 'Error' }, data: [] }));
+  test('Handles and reports 5xx error too many bytes', async () => {
+    const error: DataQueryError = {
+      message: `${LOKI_MAX_QUERY_BYTES_READ_ERROR_MSG_PREFIX} ...`,
+      status: 500,
+    };
+    const response: DataQueryResponse = {
+      state: LoadingState.Error,
+      data: [],
+      errors: [error],
+    };
+    jest.spyOn(datasource, 'runQuery').mockReturnValue(of(response));
     await expect(runSplitQuery(datasource, request)).toEmitValuesWith((values) => {
       expect(values).toHaveLength(1);
       expect(values[0]).toEqual(
-        expect.objectContaining({ error: { refId: 'A', message: 'Error' }, state: LoadingState.Streaming })
+        expect.objectContaining({
+          errors: [error],
+          state: LoadingState.Error,
+        })
       );
     });
+    // Errors are not retried
+    expect(datasource.runQuery).toHaveBeenCalledTimes(1);
+  });
+  test('Handles and reports 4xx errors', async () => {
+    const error: DataQueryError = {
+      message: 'BAD REQUEST',
+      status: 418,
+    };
+    const response: DataQueryResponse = {
+      state: LoadingState.Error,
+      data: [],
+      errors: [error],
+    };
+    jest.spyOn(datasource, 'runQuery').mockReturnValue(of(response));
+    await expect(runSplitQuery(datasource, request)).toEmitValuesWith((values) => {
+      expect(values).toHaveLength(1);
+      expect(values[0]).toEqual(
+        expect.objectContaining({
+          errors: [error],
+          state: LoadingState.Error,
+        })
+      );
+    });
+    // Errors are not retried
+    expect(datasource.runQuery).toHaveBeenCalledTimes(1);
+  });
+
+  test('Handles and reports errors (deprecated error)', async () => {
+    jest.spyOn(datasource, 'runQuery').mockReturnValue(
+      of({
+        state: LoadingState.Error,
+        error: { refId: 'A', message: `${LOKI_MAX_QUERY_BYTES_READ_ERROR_MSG_PREFIX} ...` },
+        data: [],
+        key: 'uuid',
+      })
+    );
+    await expect(runSplitQuery(datasource, request)).toEmitValuesWith((values) => {
+      expect(values).toHaveLength(1);
+      expect(values[0]).toEqual(
+        expect.objectContaining({
+          error: { refId: 'A', message: `${LOKI_MAX_QUERY_BYTES_READ_ERROR_MSG_PREFIX} ...` },
+          state: LoadingState.Error,
+        })
+      );
+    });
+    // Errors are not retried
+    expect(datasource.runQuery).toHaveBeenCalledTimes(1);
   });
 
   describe('Hidden and empty queries', () => {
     beforeAll(() => {
-      jest.spyOn(logsTimeSplit, 'splitTimeRange').mockReturnValue([]);
-      jest.spyOn(metricTimeSplit, 'splitTimeRange').mockReturnValue([]);
       jest.useFakeTimers().setSystemTime(new Date('Wed May 17 2023 17:20:12 GMT+0200'));
     });
     beforeEach(() => {
-      jest.mocked(logsTimeSplit.splitTimeRange).mockClear();
-      jest.mocked(logsTimeSplit.splitTimeRange).mockClear();
       jest.mocked(trackGroupedQueries).mockClear();
     });
     afterAll(() => {
-      jest.mocked(logsTimeSplit.splitTimeRange).mockRestore();
-      jest.mocked(metricTimeSplit.splitTimeRange).mockRestore();
       jest.useRealTimers();
     });
     test('Ignores hidden queries', async () => {
@@ -287,8 +455,6 @@ describe('runSplitQuery()', () => {
         { expr: '{a="b"}', refId: 'B' },
       ]);
       await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        expect(logsTimeSplit.splitTimeRange).toHaveBeenCalled();
-        expect(metricTimeSplit.splitTimeRange).not.toHaveBeenCalled();
         expect(trackGroupedQueries).toHaveBeenCalledTimes(1);
         expect(trackGroupedQueries).toHaveBeenCalledWith(
           {
@@ -298,16 +464,16 @@ describe('runSplitQuery()', () => {
           },
           [
             {
-              partition: [],
+              partition: expect.any(Array),
               request: {
                 ...request,
+                // Only includes queries that are not hidden
                 targets: request.targets.filter((query) => !query.hide),
               },
             },
           ],
           request,
-          new Date(),
-          { predefinedOperations: '' }
+          new Date()
         );
       });
     });
@@ -317,8 +483,6 @@ describe('runSplitQuery()', () => {
         { expr: '', refId: 'B' },
       ]);
       await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        expect(logsTimeSplit.splitTimeRange).not.toHaveBeenCalled();
-        expect(metricTimeSplit.splitTimeRange).toHaveBeenCalled();
         expect(trackGroupedQueries).toHaveBeenCalledTimes(1);
         expect(trackGroupedQueries).toHaveBeenCalledWith(
           {
@@ -328,16 +492,16 @@ describe('runSplitQuery()', () => {
           },
           [
             {
-              partition: [],
+              partition: expect.any(Array),
               request: {
                 ...request,
+                // Only includes queries with an expression
                 targets: request.targets.filter((query) => query.expr),
               },
             },
           ],
           request,
-          new Date(),
-          { predefinedOperations: '' }
+          new Date()
         );
       });
     });
@@ -444,90 +608,6 @@ describe('runSplitQuery()', () => {
     });
   });
 
-  describe('Splitting targets based on splitDuration', () => {
-    const range1h = {
-      from: dateTime('2023-02-08T05:00:00.000Z'),
-      to: dateTime('2023-02-08T06:00:00.000Z'),
-      raw: {
-        from: dateTime('2023-02-08T05:00:00.000Z'),
-        to: dateTime('2023-02-08T06:00:00.000Z'),
-      },
-    };
-    beforeEach(() => {
-      jest.spyOn(datasource, 'runQuery').mockReturnValue(of({ data: [], refId: 'A' }));
-    });
-    test('with 30m splitDuration runs 2 queries', async () => {
-      const request = {
-        targets: [{ expr: '{a="b"}', refId: 'A', splitDuration: '30m' }],
-        range: range1h,
-      } as DataQueryRequest<LokiQuery>;
-      await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        expect(datasource.runQuery).toHaveBeenCalledTimes(2);
-      });
-    });
-    test('with 1h splitDuration runs 1 queries', async () => {
-      const request = {
-        targets: [{ expr: '{a="b"}', refId: 'A', splitDuration: '1h' }],
-        range: range1h,
-      } as DataQueryRequest<LokiQuery>;
-      await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        expect(datasource.runQuery).toHaveBeenCalledTimes(1);
-      });
-    });
-    test('with 1h splitDuration and 2 targets runs 1 queries', async () => {
-      const request = {
-        targets: [
-          { expr: '{a="b"}', refId: 'A', splitDuration: '1h' },
-          { expr: '{a="b"}', refId: 'B', splitDuration: '1h' },
-        ],
-        range: range1h,
-      } as DataQueryRequest<LokiQuery>;
-      await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        expect(datasource.runQuery).toHaveBeenCalledTimes(1);
-      });
-    });
-    test('with 1h/30m splitDuration and 2 targets runs 3 queries', async () => {
-      const request = {
-        targets: [
-          { expr: '{a="b"}', refId: 'A', splitDuration: '1h' },
-          { expr: '{a="b"}', refId: 'B', splitDuration: '30m' },
-        ],
-        range: range1h,
-      } as DataQueryRequest<LokiQuery>;
-      await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        // 2 x 30m + 1 x 1h
-        expect(datasource.runQuery).toHaveBeenCalledTimes(3);
-      });
-    });
-    test('with mixed splitDuration runs the expected amount of queries', async () => {
-      const request = createRequest(
-        [
-          { expr: 'count_over_time({c="d"}[1m])', refId: 'A', splitDuration: '15m' },
-          { expr: '{a="b"}', refId: 'B', splitDuration: '15m' },
-          { expr: '{a="b"}', refId: 'C', splitDuration: '1h' },
-        ],
-        { range: range1h }
-      );
-      await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        // 4 * 15m + 4 * 15m + 1 * 1h
-        expect(datasource.runQuery).toHaveBeenCalledTimes(9);
-      });
-    });
-    test('with 1h/30m splitDuration and 1 log and 2 metric target runs 3 queries', async () => {
-      const request = createRequest(
-        [
-          { expr: '{a="b"}', refId: 'A', splitDuration: '1h' },
-          { expr: 'count_over_time({c="d"}[1m])', refId: 'C', splitDuration: '30m' },
-        ],
-        { range: range1h }
-      );
-      await expect(runSplitQuery(datasource, request)).toEmitValuesWith(() => {
-        // 2 x 30m + 1 x 1h
-        expect(datasource.runQuery).toHaveBeenCalledTimes(3);
-      });
-    });
-  });
-
   describe('Splitting targets based on resolution', () => {
     const range1d = {
       from: dateTime('2023-02-08T05:00:00.000Z'),
